/**
* HARP Version 2
*
* TODO: Test Servo Code
*       Test Watchdog Timer
*       Test Append file and f_size() macro
*/

#include "mbed.h"
#include "rtos.h"
#include "buffered_serial.h"
#include "ff.h"
#include "TMP36GZ.h"
#include "nmea_parser.h"
#include "watchdog.h"
#include "Servo.h"
#include "config.h"

Serial pc(USBTX, USBRX);
BufferedSerial gps;
AnalogIn gps_battery(p20);
AnalogIn mbed_battery(p16);
TMP36GZ temperature(p17);

NmeaParser nmea;

Semaphore parachute_sem(0);
Semaphore sd_sem(0);

typedef struct {
    char    line[80];
} gps_line;
MemoryPool<gps_line, 16> mpool_gps_line;
Queue<gps_line, 16> queue_gps_line;

typedef struct {
    char line[80];
} sensor_line;
MemoryPool<sensor_line, 16> mpool_sensor_line;
Queue<sensor_line, 16> queue_sensor_line;

void gps_thread(void const *args)
{
    char buffer[80];
    float alt, alt_prev;
    alt = alt_prev = 0;

    //DigitalOut gps_led(LED4);

    gps.baud(4800);

    while(true) {
        //gps_led = !gps_led;
        gps.get_line(buffer);
        int line_type = nmea.parse(buffer);
        //pc.puts(buffer);
        // send to log...
        gps_line *message = mpool_gps_line.alloc();
        strcpy(message->line, buffer);
        queue_gps_line.put(message);

        // test altitude direction - release parachute thread to run
        if(line_type == RMC && nmea.quality()) {
            if(UNLOCK_ON_FALL) {
                if(alt != 0) { // first time
                    alt = nmea.msl_altitude();
                } else {
                    alt_prev = alt;
                    alt = nmea.msl_altitude();
                    if(alt < alt_prev) // going down
                        parachute_sem.release(); // let the parachute code execute
                }
            } else {
                parachute_sem.release();
            }
        }
    }
}

void log_thread(const void *args)
{
    FATFS fs;
    FIL fp_gps, fp_sensor;

    DigitalOut log_led(LED3);

    f_mount(0, &fs);
    f_open(&fp_gps, "0:gps.txt", FA_OPEN_EXISTING | FA_WRITE);
    f_lseek(&fp_gps, f_size(&fp_gps));
    f_open(&fp_sensor, "0:sensors.csv", FA_OPEN_EXISTING | FA_WRITE);
    f_lseek(&fp_sensor, f_size(&fp_sensor));

    sd_sem.release(); // sd card initialized... start sensor thread
    
    while(1) {
        log_led = !log_led;
        osEvent evt1 = queue_gps_line.get(1);
        if (evt1.status == osEventMessage) {
            gps_line *message = (gps_line*)evt1.value.p;

            f_puts(message->line, &fp_gps);
            f_sync(&fp_gps);

            mpool_gps_line.free(message);
        }

        osEvent evt2 = queue_sensor_line.get(1);
        if(evt2.status == osEventMessage) {
            sensor_line *message = (sensor_line*)evt2.value.p;

            f_puts(message->line, &fp_sensor);
            f_sync(&fp_sensor);

            mpool_sensor_line.free(message);
        }
    }
}

void sensor_thread(const void* args)
{
    DigitalOut sensor_led(LED2);
    Timer t;
    float time;
    float gps_battery_voltage, mbed_battery_voltage;
    float temp;
    float distance, course_to, course;
    
    sd_sem.wait(); // wait for the sd card to initialize and open files

    if(WAIT_FOR_LOCK) {
        while(!nmea.date())  Thread::wait(100); // wait for lock
    }

    t.start(); // start timer after lock

    sensor_line *message = mpool_sensor_line.alloc();
    sprintf(message->line, "Date: %d, Time: %f\r\nGPS Time (UTC),GPS Battery(V),mbed Battery(V)", nmea.date(), nmea.utc_time());
    queue_sensor_line.put(message);
    
    message = mpool_sensor_line.alloc();
    sprintf(message->line, ",Temperature,GPS Altitude,GPS Course,Course to Target,Distance \r\n");
    queue_sensor_line.put(message);

    while(true) {
        //get sensor line memory
        sensor_led = !sensor_led;
        sensor_line *message = mpool_sensor_line.alloc();

        //timestamp
        time = nmea.utc_time();

        //gps battery
        gps_battery_voltage = gps_battery.read()*BAT_GPS_MUL;

        //mbed battery
        mbed_battery_voltage = mbed_battery.read()*BAT_MBED_MUL;

        //temperature
        temp = temperature.sample_f();
        
        //gps
        distance = nmea.calc_dist_to_km(target_lat, target_lon);
        course_to = nmea.calc_initial_bearing(target_lat, target_lon);
        course = nmea.track();

        sprintf(message->line, "%f,%f,%f,%f,%f,%f,%f,%f\r\n", time,gps_battery_voltage,mbed_battery_voltage,temp,nmea.calc_altitude_ft(),course,course_to,distance);
        queue_sensor_line.put(message);
        
        Thread::wait(100);
    }
}

void parachute_thread(const void *args)
{
    DigitalOut left_turn(LED4);
    DigitalOut right_turn(LED1);
    
    // servos ////////////////////////// NOT TESTED!!! ///////////////////////////
    Servo left_s(p21);
    Servo right_s(p22);

    left_s.calibrate_max(0.0007);
    left_s.calibrate_min(-0.0014);
    right_s.calibrate(0.0009);
    
    left_s = 0.0;
    right_s = 0.0;
    ////////////////////////////////////////////////////////////////////////////////

    right_turn = 1;
    Thread::wait(400);
    left_turn = 1;
    Thread::wait(400);
    while(true) {
        right_turn = left_turn = 0;
        parachute_sem.wait();

        float distance = nmea.calc_dist_to_km(target_lat, target_lon);

        if(distance < distance_fudge_km)
            continue; // dont do anything

        float course = nmea.track();
        float course_to = nmea.calc_initial_bearing(target_lat, target_lon);
        float course_diff = course_to - course;
        
        if(course == 0.0) // not moving fast enough
            continue; // do nothing

        if(course_diff < course_fudge && course_diff > neg_course_fudge) {
            right_turn = left_turn = 1;
            Thread::wait(400);
            continue; // don't do anything
        } else if(course_diff > 180.0 || course_diff < 0.0) {
            left_turn = 1;
            right_s = 0.0;                                      // NOT TESTED!!!
            left_s = 1.0;
            Thread::wait(400); // turn left
            left_s = 0.0;
        } else {
            right_turn = 1;
            left_s = 0.0;
            right_s = 1.0;
            Thread::wait(400); // turn righ
            right_s = 0.0;
        }
    }
}

int main()
{
    pc.baud(9600);
    Thread thread(gps_thread, NULL, osPriorityHigh);
    Thread thread2(log_thread, NULL, osPriorityNormal);
    Thread thread3(sensor_thread, NULL, osPriorityNormal);
    Thread thread4(parachute_thread, NULL, osPriorityRealtime);
    
    Watchdog wdt; // NOT TESTED!!!
    
    wdt.kick(2.0);

    while(true) {
        wdt.kick();
        Thread::wait(500);
    }
}